#ifndef LocalVField_h
#define LocalVField_h

#include "capd/capdlib.h"

#include "utils.h"

using namespace std;
using namespace capd;
using namespace capd::matrixAlgorithms;
using capd::autodiff::Node;

//////////////////////////////////
// class: LocalVField
//
// Purpose: The aim of the class is to consider a vector field in local coordinates in a neighbourhood of
// a fixed point. 
//
// The local coordinates "p" are: 
//    q = fixedPoint + C*psi(p)
// where "q" are the original coordinates of the ODE.
//
// Members:
//         F    - the vector field in original coordinates
//   fixedPoint - The fixed point of the flow. i.e: Phi(T,fixedPoint)=fixedPoint
//         C    - Matrix which gives the linear change of coordinates.
//       psi    - function that includes the nonlinear part of change of coordinates
//
// Derivation of the local vector field:
//    q = fixedPoint + C*psi(p)=Psi(p)
//    p = psi^(-1)(C^(-1)*(q - fixedPoint))
//
// The local vector field is:
//    Floc(p) = (DPsi(p))^(-1)*F(Psi(p))
//    DFloc(p)= (DPsi(p))^(-1)(DF(Psi(p))*DPsi(p)-D^2Psi(p)*Floc(p))
// where
//    DPsi(p)=CDpsi(p)
//    D^2 Psi(p)=CD^2 psi(p)

class LocalVField
{
public:
	IMap F;
	IVector fixedPoint;
	IMatrix C;
    IMap psi;
	
	LocalVField(IMap F, IVector fixedPoint, IMatrix C, IMap psi);
    
    IVector Psi(const IVector &p) const;
    IMatrix DPsi(const IVector &p) const;
    
    IVector image(const IVector &p) const;
    IVector operator()(const IVector &p) const;

	IMatrix derivative(const IVector &p) const;
	IMatrix operator[](const IVector &p) const;
    
    IMatrix D2psi(const IVector &p) const;
    
    C0HORect2Set set_init_cond(const IVector &Bloc) const;
    
};

#endif
