#include <iostream>

#include "capd/capdlib.h"
#include "capd/mpcapdlib.h"

#include "utils.h"
#include "LocalVField.h"
#include "conecond.h"
#include "ThreeBodyProblemVectorField.h"

using namespace std;
using namespace capd;
using namespace capd::matrixAlgorithms;
using capd::autodiff::Node;

void TBP_homoclinic_proof()
{
    cout.precision(9);
    
    interval mu = 0.004253863522+power(interval(10.0),-10)*interval(-1.,1.);
    IVector mu_bound(2);
    mu_bound[0] = mu.left();
    mu_bound[1] = mu.right();
    
    // we define the vector field of the Three Body Problem for whole range of parameters
    ThreeBodyProblemVectorField TBP(mu);
    cout << "Parameters of the Three Body Problem for whole interval of mu:" << endl;
    TBP.print_param();
    
    ThreeBodyProblemVectorField TBP_temp(mu);
    
    //we compute parameters of the vector field for mu_left and mu_right
    for(int i=0;i<2;i++)
    {
        TBP_temp.change_mu(mu_bound[i]);
        cout << "Parameters of the Three Body Problem for mu[" << i << "]:" << endl;
        TBP_temp.print_param();
    }
    
    cout << "------------------------------------------------------------------------" << endl << endl;
    
    // we set the parameters for finding the enlosure of the unstable manifold
    // Remark: values of the parameters are the same for each mu
    
    // r_u (r_s) is the radius of the domain (codomain) of an unstable manifold
    // alfa,beta, ni_alpha, nie_beta are parameters of the cone conditions
    // Remark: alpha and beta must satisfy alpha>beta
    
    interval r_u = pow(10.0,-6.0);
    interval beta = pow(10.0,-2.0);
    interval r_s = beta*r_u;
    interval alpha_beta_coef = 10000;
    interval alpha = alpha_beta_coef*beta;
    
    if((alpha>beta)==0 ) abort("Condition alpha>beta not satisfied. ABORTING program. ");
    
    interval ni_alpha = 2.8;
    interval ni_beta = 0.02;
    
    // u is the dimension of the unstable manifold
    // n is the dimension of the vector field
    int u = 1;
    int n = 4;
    IVector zero(n);
    
    // U is a product of [-r_u,r_u]^u x [-r_s,r_s]^(n-u)
    // on this set we will later check cone codnitions
    // Remark: observe that U contains N(0,r_u,r_s)=B_u(r_u) x B_s(r_s)
    IVector U(n);
    
    for(int i=0;i<u;i++)
        U[i] = interval(-1.,1.)*r_u;
    for(int i=u;i<n;i++)
        U[i] = interval(-1.,1.)*r_s;
    
    cout << "Parameters for the cone conditions for whole interval of mu:" << endl;
    cout << "r_u:" << r_u << endl;
    cout << "r_s:" << r_s << endl << endl;
    
    cout << "alpha:" << alpha << endl;
    cout << "beta:" << beta << endl;
    cout << "ni_alpha:" << ni_alpha << endl;
    cout << "ni_beta:" << ni_beta << endl << endl;
    cout << "U:" << U << endl << endl;
    cout << "------------------------------------------------------------------------" << endl << endl;

    ////////////////////////////////////////////////////////////////////
    // we check that the local vector field of the Three Body Problem satisfies cone conditions on a set N(0,r_u,r_s)
    
    // to check cone conditions we divide set mu into smaller parts and check the cone condition for each vector field separately
    
    int number_of_slices_U = 1000;
    
    cout << "Checking cone conditions for given parameters for whole interval of mu..." << endl << endl;
    cout << " N_U: " << number_of_slices_U << endl << endl;
    
    interval mu_fragment;
    
    IMatrix DFlocU(n,n);
    IVector U_fragment=U;
    
    // computation of the enclosure of matrix DFloc(U)
    // Remark: to make the enclosure smaller we subdivide U into smaller sets and compute the enlosure on each subset separately
    DFlocU = (TBP.Floc)[zero];
        
    for(int j=0;j<number_of_slices_U;j++)
    {
        U_fragment[0] = part( U[0], number_of_slices_U, j );
        intervalHull(TBP.Floc[U_fragment],DFlocU);
    }
    
    cout << "[DFloc(U)] for whole interval of mu:" << endl << DFlocU << endl << endl;
        
    // we check if DFloc satisfies proper cone conditions in U
    
    if(satisfiesConeConditions( alpha, u, DFlocU, ni_alpha) ==0 ) abort("unable to verify the cone conditions for (Q_alpha,m_alpha). ABORTING program. ");
    if(satisfiesConeConditions( beta, u, DFlocU, ni_beta) ==0 ) abort("unable to verify the cone conditions for (Q_beta,m_beta). ABORTING program. ");

    cout << "Local Vector Field satisfies proper cone conditions for given parameters for whole interval of mu." << endl << endl;
    cout << "------------------------------------------------------------------------" << endl << endl;
    
    //----------------------------------------------------------------//
    
    ////////////////////////////////////////////////////////////////////
    // we find the (tihghter) enclosure on the unstable manifolrd for mu_left and mu_right
    // we want to make parametr beta smaller but without changing parameter y_u_radius
    
    interval r_u_bound = r_u;
    interval beta_bound = 5*pow(10,-3)*beta;
    interval r_s_bound = beta_bound*r_u_bound;
    interval alpha_beta_coef_bound = alpha_beta_coef;
    interval alpha_bound = alpha_beta_coef_bound*beta_bound;
    
    if((alpha_bound>beta_bound)==0 ) abort("condition alpha>beta not satisfied. ABORTING program. ");
    
    interval ni_alpha_bound = ni_alpha;
    interval ni_beta_bound = ni_beta;
    
    IVector U_bound(n);
    
    for(int i=0;i<u;i++)
        U_bound[i] = interval(-1.,1.)*r_u_bound;
    for(int i=u;i<n;i++)
        U_bound[i] = interval(-1.,1.)*r_s_bound;

    
    cout << "Parameters for the cone conditions for mu_left and mu_right:" << endl;
    cout << "r_u:" << r_u_bound << endl;
    cout << "r_s:" << r_s_bound << endl << endl;
    
    cout << "alpha:" << alpha_bound << endl;
    cout << "beta:" << beta_bound << endl;
    cout << "ni_alpha:" << ni_alpha_bound << endl;
    cout << "ni_beta:" << ni_beta_bound << endl << endl;
    cout << "U:" << U_bound << endl << endl;
    cout << "------------------------------------------------------------------------" << endl << endl;
    
    IVector U_bound_fragment=U_bound;
    number_of_slices_U = 100000;
    
    cout << "Checking cone conditions for given parameters for mu_left and mu_right..." << endl << endl;
    cout << "N_U: " << number_of_slices_U << endl << endl;
    
    for(int i=0;i<2;i++)
    {
        TBP_temp.change_mu(mu_bound[i]);
        
        // computation of the enclosure of matrix DFloc(U)
        // Remark: to make the enclosure smaller we subdivide U into smaller sets and compute the enlosure on each subset separately
        DFlocU = (TBP_temp.Floc)[zero];
        
        for(int j=0;j<number_of_slices_U;j++)
        {
            U_bound_fragment[0] = part( U_bound[0], number_of_slices_U, j );
            intervalHull(TBP_temp.Floc[U_bound_fragment],DFlocU);
        }
        
        cout << "[DFloc(U)] for mu[" << i << "]:" << endl << DFlocU << endl << endl;
        
        // we check if DFloc satisfies proper cone conditions in U
        
        if(satisfiesConeConditions( alpha_bound, u, DFlocU, ni_alpha_bound) ==0 ) abort("unable to verify the cone conditions for (Q_alpha,m_alpha). ABORTING program. ");
        if(satisfiesConeConditions( beta_bound, u, DFlocU, ni_beta_bound) ==0 ) abort("unable to verify the cone conditions for (Q_beta,m_beta). ABORTING program. ");
    }
    
    cout << "Local Vector Field satisfies proper cone conditions for given parameters for mu_left and mu_right." << endl << endl;
    cout << "------------------------------------------------------------------------" << endl << endl;
    
    
    //----------------------------------------------------------------//
    
    ////////////////////////////////////////////////////////////////////
    // we compute Bloc and B - the sets throgh which passes the unstable manifold in local and in original coordinates
    // Remark: Bloc is the same for each mu, sice the local vector field for each mu satisfies the same cone conditions
    
    // parameter y_u_radius is a radius of domain of y_u
    interval y_u_radius = r_u *sqrt(1.-power(alpha/beta,-2));
    
    cout << "Radius of domain of y_u for whole interval mu:" << y_u_radius << endl << endl;
    
    interval y_u_radius_bound = r_u_bound *sqrt(1.-power(alpha_bound/beta_bound,-2));
    
    cout << "Radius of domain of y_u for mu_left nad mu_right:" << y_u_radius_bound << endl << endl;
    
    // Bloc is the set through which passes unstable manifold in local coordinates for whole interval of mu
    
    IVector Bloc(n);
    
    for(int i=0;i<u;i++)
        Bloc[i] = y_u_radius.left();
    for(int i=u;i<n;i++)
        Bloc[i] = interval(-1.,1.)*r_s;
    
    cout << "Bloc for whole interval of mu:" << Bloc << endl << endl;
    cout << "After transporting the set Bloc to original coordinates we get set B of the form: L1 + " << TBP.C*TBP.psi(Bloc) << endl << endl;
    
    // Bloc_bound is the set through which passes unstable manifold in local coordinates for mu_right and mu_left
    IVector Bloc_bound(n);
    
    for(int i=0;i<u;i++)
        Bloc_bound[i] = y_u_radius_bound.left();
    for(int i=u;i<n;i++)
        Bloc_bound[i] = interval(-1.,1.)*r_s_bound;
    
    cout << "Bloc for mu_left and mu_right:" << Bloc_bound << endl << endl;
    for(int i=0;i<2;i++)
    {
        TBP_temp.change_mu(mu_bound[i]);
        
        cout << "After transporting the set Bloc to original coordinates for mu[" << i << "] we get set B of the form: L1 + " << TBP_temp.C*TBP_temp.psi(Bloc_bound) << endl << endl;
    }

    cout << "------------------------------------------------------------------------" << endl << endl;
    
    
    ////////////////////////////////////////////////////////////////////
    
    //----------------------------------------------------------------//
    
    ////////////////////////////////////////////////////////////////////
    // we compute the intersection of B with Poincare section for mu_left and mu_right
    
    // B_init is the domain of the Poincare Map - it is a set Bloc_bound transformed to original cordinates
    
    int order=10;
    ICoordinateSection section(4,1);
    
    IVector PX_Poincare(2);
    
    cout << "Computing the intersection of the unstable mniafold with Poincare section for mu_left and mu_right..." << endl << endl;
    
    for(int i=0;i<2;i++)
    {
        TBP_temp.change_mu(mu_bound[i]);
        
        C0HORect2Set B_init = TBP_temp.Floc.set_init_cond(Bloc_bound);
        IOdeSolver solver(TBP_temp.F,order);
        IPoincareMap P(solver,section);
        ITimeMap PHI(solver);
        
        IVector p = P(B_init);
        cout << "intersection of the manifold with {Y=0} for mu[" << i << "]:" << endl;
        cout << p << endl << endl;
        cout << "-------------------------------------" << endl << endl;
        PX_Poincare[i] = p[2];
    }
    
    cout << "Checking if the computed intersections have different signs on the third coordinate (PX)... " << endl << endl;
    
    if(PX_Poincare[0]*PX_Poincare[1]<0)
        cout << "The computed intersections have different signs on the third coordinate" << endl << endl;
        else
        abort("unable to verify if the computed intersections have different signs on the third coordinate. ABORTING program. ");
    cout << "-------------------------------------" << endl << endl;
    
    // we check if the Poincare map is well defined for whole range of parameters of mu
    
    int number_of_slices_mu = 10;
    
    cout << "Checking if the Poincare map is well defined for whole interval of mu..." << endl << endl;
    cout<< "N_mu: " << number_of_slices_mu <<  endl << endl;
    
    for(int i=0;i<number_of_slices_mu;i++)
    {
        TBP_temp.change_mu(part(mu,number_of_slices_mu,i));
        
        C0HORect2Set B_init = TBP_temp.Floc.set_init_cond(Bloc);
        IOdeSolver solver(TBP_temp.F,order);
        IPoincareMap P(solver,section);
        ITimeMap PHI(solver);
        
        IVector p = P(B_init);
    }
    
    cout << "Poincare map is well defined for whole range of parameters mu" << endl << endl;
    cout << "-------------------------------------" << endl << endl;
    
    ////////////////////////////////////////////////////////////////////
    
    //----------------------------------------------------------------//
    
    
    cout << "End of proof" << endl << endl;
}

int main(int argc, char* argv[])
{
    time_t before,after;
    before=time(NULL);
	
  	cout.precision(18);
	try
	{
        TBP_homoclinic_proof();

	}
	catch(exception& e)
  	{
    		cout << "\n\nException caught: "<< e.what() << endl;
  	}
    
	after=time(NULL);
    cout << "time: " << difftime(after,before) << " s. " << endl;
  	return 0;
} 
