#include "ThreeBodyProblemVectorField.h"

// string contaning definition of a vector field in The Three Body Problem
string vField="par:mu;var:X,Y,PX,PY;fun:PX+Y,PY-X,PY-(1-mu)*(X-mu)*((X-mu)*(X-mu)+Y*Y)^(-1.5)-mu*(X-mu+1)*((X-mu+1)*(X-mu+1)+Y*Y)^(-1.5),-PX-(1-mu)*Y*((X-mu)*(X-mu)+Y*Y)^(-1.5)-mu*Y*((X-mu+1)*(X-mu+1)+Y*Y)^(-1.5);";

// string contaning definition of a function psi
string psiformula="var:x,y1,y2,y3;fun:x-y1*(-0.8853994638241132*x+0.6351923719779123*x^2)-y2*(1.4409405088342198*x-0.6232244954364759*x^2)-y3*(1.2193508824506356*x-4.8745113996400465*x^2),y1-0.4426997319120566*x^2+0.2117307906593041*x^2*x,y2+0.7204702544171099*x^2-0.2077414984788253*x^2*x,y3+0.6096754412253178*x^2-1.6248371332133488*x^2*x;";


void ThreeBodyProblemVectorField::change_mu(interval mu_)
{
    mu=mu_;
    init_F();
    init_L1();
    init_C();
    init_Floc();
}

void ThreeBodyProblemVectorField::init_F()
{
    ////////////////////////////////////////////////////////////////////
    // definition of the vector field F
    
    // vector Field of The Three Body Problem, parameter 2 is needed to store the derivatives of F up to order 2
    F = IMap(vField,2);
    F.setParameter("mu",mu);
}

void ThreeBodyProblemVectorField::init_L1()
{
    ////////////////////////////////////////////////////////////////////
    // computation of the equilibrium L1
    
    // the initial guess of L1, which has the form (xL1,0,0,xL1)
    // numerical computations suggest that xL1 is close to -0.9
    IVector L1_guess(4);
    L1_guess[0] = -0.9;
    L1_guess[1] = 0.0;
    L1_guess[2] = 0.0;
    L1_guess[3] = -0.9;
    
    // equilibrium point L1 in original coordinates
    // Remark: this is a rigorous enclosure of family of equilibrium points L1^mu for a given interval of parameters mu
    // L1 is found using Interval Newton Method
    L1 = FixedPoint(F,L1_guess);
    
    // From theoretical arguments we know that:
    L1[1] = 0.0;
    L1[2] = 0.0;
    // From theoretical arguments we know that:
    interval xL1;
    intersection(L1[0],L1[3],xL1);
    L1[0] = xL1;
    L1[3] = xL1;
}

void ThreeBodyProblemVectorField::init_C()
{
    ////////////////////////////////////////////////////////////////////
    // computation of values lambda and v which define the eigenvalues of the jacobian of the vector field at L1
    interval xL1 = L1[0];
    
    interval gamma = xL1 + 1. - mu;
    interval c2 = mu/(gamma*gamma*gamma)+(1.-mu)/((1.-gamma)*(1.-gamma)*(1.-gamma));
    
    lambda = sqrt ((c2 - 2. + sqrt (9. * (c2*c2) - 8. * c2)) / 2.);
    v = sqrt ((-c2 +2. +sqrt (9. * (c2*c2) - 8. * c2)) / 2.);
    
    ////////////////////////////////////////////////////////////////////
    
    //----------------------------------------------------------------//
    
    ////////////////////////////////////////////////////////////////////
    // computation of matrix C which define the linear coordinate change at L1
    
    interval s1 = sqrt (2. * lambda * ((4. + 3. * c2) * (lambda*lambda) + 4. + 5. * c2 - 6. * (c2*c2)));
    interval s2 = sqrt (v * ((4. + 3. * c2) * (v*v) - 4. - 5. * c2 + 6. * (c2*c2) ));
    
    C = IMatrix(4,4);
    
    C[0][0] = 2. * lambda / s1;
    C[0][1] = -2. * lambda / s1;
    C[0][2] = 0.;
    C[0][3] = 2. * v / s2;
    C[1][0] = ((lambda*lambda) - 2. * c2 - 1.) / s1;
    C[1][1] = ((lambda*lambda) - 2. * c2 - 1.) / s1;
    C[1][2] = (-(v*v) - 2. * c2 - 1.) / s2;
    C[1][3] = 0.;
    C[2][0] = ((lambda*lambda) + 2. * c2 + 1.) / s1;
    C[2][1] = ((lambda*lambda) + 2. * c2 + 1.) / s1;
    C[2][2] = (-(v*v) + 2. * c2 + 1.) / s2;
    C[2][3] = 0.;
    C[3][0] = ((lambda*lambda*lambda) + (1. - 2. * c2) * lambda) / s1;
    C[3][1] = (-(lambda*lambda*lambda) - (1. - 2. * c2) * lambda) / s1;
    C[3][2] = 0.;
    C[3][3] = (-(v*v*v) + (1. - 2. * c2) * v) / s2;
}

void ThreeBodyProblemVectorField::init_psi()
{
    ////////////////////////////////////////////////////////////////////
    // definition of the nonlinear coordinate change
    
    // function psi, parameter 2 is needed to store the derivatives of psi up to order 2
    psi = IMap(psiformula,2);
    
    ////////////////////////////////////////////////////////////////////
}

void ThreeBodyProblemVectorField::init_Floc()
{
    ////////////////////////////////////////////////////////////////////
    // definition of the local vector field
    Floc = LocalVField(F, L1, C, psi);
}

ThreeBodyProblemVectorField::ThreeBodyProblemVectorField(interval mu_):Floc(F, L1, C, psi)
{
    mu=mu_;
    init_F();
    init_L1();
    init_C();
    init_psi();
    init_Floc();
}

void ThreeBodyProblemVectorField::print_param()
{
    cout << "mu:" << mu << endl << endl;
    cout << "Equilibrium point L1:" << endl << L1 << endl << endl;
    cout << "lambda:" << endl << lambda << endl;
    cout << "v:" << endl << v << endl << endl;
    cout << "Matrix C" << endl << C << endl;
    
    //Jacobi matrix of local vector field at fixed point = 0
    IVector zero(4);
    cout << "DFloc(0)" << endl << Floc[zero] << endl << endl;
}

