#include "conecond.h"

////////////////////////////////////////////////////
// Function: GershgorinEigenvalueBound()
// 
// Purpose: computes the bound on the eigenvalues using the Gershgorin theorem.
// The function returns an interval, which contains all eigenvalues of A.
// The function is meant to be applied in a setting where all eigenvalues 
// have the same sign, and we intend to show that they are separated from zero. 
//
// Note: To apply this function, the matrix A needs to be in the form that is close to
// diagonal. The entries outside of the diagonal need to be small in order to obtain 
// sharp estimates.
interval GershgorinEigenvalueBound(const IMatrix &A)
{
   int n=A.numberOfRows();
   interval bound = A[0][0];
   for(int i=0;i<n;i++)
   {
      interval h(0.0);
      for(int k=0;k<n;k++)
      {
         if(!(i==k)) h=h+abs(A[i][k]);
      }
      bound = intervalHull(bound, A[i][i] + intervalHull(-h,h));
   }
   return bound;
}

////////////////////////////////////////////////////
// Function: isPositiveDefinite()
//
// Purpose: The function checks whether A is strictly positive definite.
//
// Note: The function checks if all eigenvalues are positive. To apply this function, 
// the matrix A needs to be in the form that is close to diagonal. The entries outside 
// of the diagonal need to be small.
// If the given matrix A is close to symmetric, but the entries outside the diagonal
// are large, instead of the matrix A, we can consider matrix B
// obtained from matrix A by changing pairs of symmetric entries (a,a) to
// (a-a.mid,a+a.mid). It can be easily show that for any vector x the equality
// x^TAx=x^TBx
// is satisfied and therefore 
// the above condition
bool isPositiveDefinite(const IMatrix &A)
{
	int n=A.numberOfColumns();
	IMatrix B=A;
	interval a;
	
	if(n>1)
	{
		for(int i=0;i<n;i++)
			for(int k=i+1;k<n;k++)
			{
				a=A[i][k].mid();
				B[i][k]=B[i][k]-a;
				B[k][i]=B[k][i]+a;
			}
		if(GershgorinEigenvalueBound(B)>0.0) return 1;
	}
   if(GershgorinEigenvalueBound(A)>0.0) return 1;
   return 0;
}

////////////////////////////////
// Function: satisfiesConeConditions()
//
// The purpose of this function is to check if the flow phi_t of the vector field
// satisfies cone conditions for (Q,m=1+2t*ni) on a given set U. The enclosure of the derivative of the vector field on U has a form
// A  E1
// E2 B,
// where A, E1, E2, B are interval matrices of dimensions u x u, u x (n-u), (n-u) x u, and (n-u) x (n-u).
// The mapping Q is of the form
// Q(x,y)=alpha^2*||x||^2-||y||^2.
//
// Note: the intervals e1 and e2 given as arguments of a function are the euclidean norms of matrices E1 and E2.
bool satisfiesConeConditions(const interval &alpha,	const IMatrix &A,const IMatrix &B, const interval &e1,const interval &e2,const interval &ni)
{
	int u=A.numberOfColumns();
    IMatrix Id_u = IdMatrix(u);
	if(isPositiveDefinite(A-0.5*(e1+e2/(alpha^2)+2.0*ni)*Id_u)==0) return 0;
	
	int s=B.numberOfColumns();
	IMatrix Id_s = IdMatrix(s);
	return isPositiveDefinite(-B+0.5*(-e2-e1*(alpha^2)+2.0*ni)*Id_s);
}

////////////////////////////////
// Function: satisfiesConeConditions()
//
// The purpose of this function is to check if the flow phi_t of the vector field
// satisfies cone conditions for (Q,m=1+2t*ni). DF is the interval enclosure of a derivative of a vector
// field on the given set U.
// The mapping Q is of the form
// Q(x,y)=alpha^2*||x||-||y||^2.
// u - is the dimension of an unstable manifold (so also the dimension of x).
bool satisfiesConeConditions(const interval &alpha,
								int u, const IMatrix &DF, const interval &ni)
{
	int n=DF.numberOfColumns();
	IMatrix A(u,u);
	IMatrix B(n-u,n-u);
	IMatrix E1(u,n-u);
	IMatrix E2(n-u,u);
	
	for(int i=0;i<u;i++)
		for(int j=0;j<u;j++) A[i][j]=DF[i][j];
	for(int i=0;i<n-u;i++)
		for(int j=0;j<n-u;j++) B[i][j]=DF[u+i][u+j];
	for(int i=0;i<u;i++)
		for(int j=0;j<n-u;j++) E1[i][j]=DF[i][u+j];
	for(int i=0;i<n-u;i++)
		for(int j=0;j<u;j++) E2[i][j]=DF[u+i][j];
	
	
	interval e1=euclNorm(E1);
	interval e2=euclNorm(E2);
	
	return satisfiesConeConditions(alpha,A,B,e1,e2,ni);
}
